@@ -1,6 +1,5 @@
-from .worstcase import Config as config
 from .worstcase import Derivative as derive
 from .worstcase import Parameter as param
 from .worstcase import Unit as unit
 
-__all__ = ["config", "param", "derive", "unit"]
+__all__ = ["param", "derive", "unit"]
