@@ -1,13 +1,13 @@
 from enum import Enum, auto
 from inspect import Signature
 from itertools import product
-from types import SimpleNamespace
+from warnings import warn
 
-import networkx as nx  # type: ignore
-from pint import Quantity, UnitRegistry  # type: ignore
-from pyDOE import lhs  # type: ignore
+import networkx as nx
+import numpy as np
+from pint import Quantity, UnitRegistry
+from pyDOE import lhs
 
-Config = SimpleNamespace(n=5000, sigfig=4)
 Unit = UnitRegistry()
 
 
@@ -32,7 +32,7 @@ class AbstractParameter:
 
 
 class Parameter(AbstractParameter):
-    def __init__(self, nom, lb, ub, tag):
+    def __init__(self, nom, lb, ub, tag, sigfig):
         nom = nom if isinstance(nom, Quantity) else nom * Unit([])
         lb = lb if isinstance(lb, Quantity) else lb * Unit([])
         ub = ub if isinstance(ub, Quantity) else ub * Unit([])
@@ -44,18 +44,19 @@ class Parameter(AbstractParameter):
         self.lb = lb  # lower bound quantity
         self.ub = ub  # upper bound quantity
         self.tag = tag  # string identifier
+        self.sigfig = sigfig  # string significant digits
 
     @staticmethod
-    def byrange(nom, lb, ub, tag=""):
-        return Parameter(nom, lb, ub, tag)
+    def byrange(nom, lb, ub, tag="", sigfig=4):
+        return Parameter(nom, lb, ub, tag, sigfig)
 
     @staticmethod
-    def bytol(nom, tol, rel, tag=""):
+    def bytol(nom, tol, rel, tag="", sigfig=4):
         tol = nom * tol if rel else tol
-        return Parameter(nom, nom - tol, nom + tol, tag)
+        return Parameter(nom, nom - tol, nom + tol, tag, sigfig)
 
     def __repr__(self):
-        pretty = "0.{sigfig}G~P".format(sigfig=Config.sigfig)
+        pretty = "0.{sigfig}G~P".format(sigfig=self.sigfig)
         tag = "{tag}: ".format(tag=self.tag) if self.tag else ""
         nom = ("{nom:" + pretty + "} (nom), ").format(nom=self.nom.to_compact())
         lb = ("{lb:" + pretty + "} (lb), ").format(lb=self.lb.to_compact())
@@ -95,35 +96,45 @@ class Parameter(AbstractParameter):
 class By(Enum):
     EV = auto()
     MC = auto()
+    RSS = auto()
 
 
 class Derivative(AbstractParameter):
-    def __init__(self, func, method, tag, *args, **kwargs):
+    def __init__(self, func, method, tag, sigfig, n, *args, **kwargs):
         self.func = func
         self.method = method
         self.tag = tag
+        self.sigfig = sigfig
+        self.n = n
 
         funcsig = Signature.from_callable(func)
         binding = funcsig.bind_partial(*args, **kwargs)
         self.kwargs = binding.arguments
 
     @staticmethod
-    def byev(*args, tag="", **kwargs):
+    def byev(*args, tag="", sigfig=4, n=None, **kwargs):
         def decorator(func):
-            return Derivative(func, By.EV, tag, *args, **kwargs)
+            return Derivative(func, By.EV, tag, sigfig, n, *args, **kwargs)
 
         return decorator
 
     @staticmethod
-    def bymc(*args, tag="", **kwargs):
+    def bymc(*args, tag="", sigfig=4, n=1000, **kwargs):
         def decorator(func):
-            return Derivative(func, By.MC, tag, *args, **kwargs)
+            return Derivative(func, By.MC, tag, sigfig, n, *args, **kwargs)
+
+        return decorator
+
+    @staticmethod
+    def byrss(*args, tag="", sigfig=4, n=None, **kwargs):
+        def decorator(func):
+            return Derivative(func, By.RSS, tag, sigfig, n, *args, **kwargs)
 
         return decorator
 
     @property
     def nom(self):
-        return self.derive().nom
+        return eval_nominal(self.graph(), self)[1]
 
     @property
     def lb(self):
@@ -136,7 +147,7 @@ class Derivative(AbstractParameter):
     def __repr__(self):
         return self.derive().__repr__()
 
-    def __call__(self, *args, tag=None, **kwargs):
+    def __call__(self, *args, tag=None, sigfig=None, n=None, **kwargs):
 
         # If args/kwargs form a complete binding with no AbstractParameters
         # then return a simple call to the underlying function.
@@ -164,7 +175,9 @@ class Derivative(AbstractParameter):
         new_kwargs.update(binding.arguments)
 
         tag = self.tag if tag is None else tag
-        return Derivative(self.func, self.method, tag, **new_kwargs)
+        sigfig = self.sigfig if sigfig is None else sigfig
+        n = self.n if n is None else n
+        return Derivative(self.func, self.method, tag, sigfig, n, **new_kwargs)
 
     def graph(self, ss=None):
 
@@ -176,11 +189,10 @@ class Derivative(AbstractParameter):
         # - each node is an AbstractParameter
         # - each node has a "latest" field to hold the latest
         #   AbstractParameter assignment which will be used in evaluation
-        # - each directed edge has a "method" field (By.MC or By.EV)
         graph = nx.DiGraph()
         graph.add_node(self, latest=self)
         for pred in preds:
-            graph.add_edge(pred, self, method=self.method)
+            graph.add_edge(pred, self)
             graph = nx.compose(graph, pred.graph())
 
         # A sensitivity study is provided by the "ss" argument, a list of
@@ -202,7 +214,9 @@ class Derivative(AbstractParameter):
 
             for node in graph.nodes:
                 if is_param(node) and node not in ss_params:
-                    latest = Parameter(node.nom, node.nom, node.nom, node.tag)
+                    latest = Parameter(
+                        node.nom, node.nom, node.nom, node.tag, node.sigfig
+                    )
                     graph.nodes[node]["latest"] = latest
 
         return graph
@@ -241,12 +255,14 @@ class Derivative(AbstractParameter):
             for node in eval_nodes:
                 ancestors = nx.ancestors(graph, node)
                 if not any([anc in eval_nodes for anc in ancestors]):
-                    # Evaluate using the current nodes method (EV or MC).
+                    # Evaluate using the current nodes method (EV, MC, or RSS).
                     # This will overrule any depending node's method.
                     if node.method is By.EV:
                         param = extreme_value(graph, node)
-                    else:
+                    elif node.method is By.MC:
                         param = monte_carlo(graph, node)
+                    else:
+                        param = root_sum_square(graph, node)
 
                     # Update the current node with the latest evaluation.
                     graph.nodes[node]["latest"] = param
@@ -326,14 +342,14 @@ def extreme_value(graph, eval_node):
         lbmin = result if result < lbmin else lbmin
         ubmax = result if result > ubmax else ubmax
 
-    return Parameter(nom, lbmin, ubmax, eval_node.tag)
+    return Parameter(nom, lbmin, ubmax, eval_node.tag, eval_node.sigfig)
 
 
 def monte_carlo(graph, eval_node):
     params, nom = eval_nominal(graph, eval_node)
 
     # Run n Monte Carlo evaluations with Latin Hypercube Sampling.
-    matrix = lhs(len(params), samples=Config.n)
+    matrix = lhs(len(params), samples=eval_node.n)
     results = []
     for row in matrix:
         eval_init = {}
@@ -342,4 +358,34 @@ def monte_carlo(graph, eval_node):
             eval_init[p] = x * (latest.ub - latest.lb) + latest.lb
         results.append(eval_graph(graph, eval_node, eval_init))
 
-    return Parameter(nom, min(results), max(results), eval_node.tag)
+    return Parameter(nom, min(results), max(results), eval_node.tag, eval_node.sigfig)
+
+
+def root_sum_square(graph, eval_node):
+    params, nom = eval_nominal(graph, eval_node)
+
+    # For each parameter, calculate the result using both the lower and upper
+    # bounds while holding all other parameters at their nominal value.
+    lbs = []
+    ubs = []
+    for pvaried in params:
+        eval_init = {p: graph.nodes[p]["latest"].nom for p in params - {pvaried}}
+        eval_init[pvaried] = graph.nodes[pvaried]["latest"].lb
+        ret_lb = eval_graph(graph, eval_node, eval_init).m
+        eval_init[pvaried] = graph.nodes[pvaried]["latest"].ub
+        ret_ub = eval_graph(graph, eval_node, eval_init).m
+        ubs.append(max([ret_lb, ret_ub]))
+        lbs.append(min([ret_lb, ret_ub]))
+
+    # Define a metric to detect an asymmetric result and warn the user.
+    ub_deviation = np.array(ubs) - nom.m
+    lb_deviation = nom.m - np.array(lbs)
+    mean_deviation = (ub_deviation + lb_deviation) / 2
+    max_deviation = np.maximum(ub_deviation, lb_deviation)
+    if np.any(max_deviation > mean_deviation * 1.05):
+        warn("RSS method is not recommended for asymmetric derived parameters.")
+
+    # Compute the RSS (root-sum-square).
+    ub = nom.m + np.sqrt(np.sum(np.square(np.array(ubs) - nom.m)))
+    lb = nom.m - np.sqrt(np.sum(np.square(np.array(lbs) - nom.m)))
+    return Parameter(nom, lb * nom.u, ub * nom.u, eval_node.tag, eval_node.sigfig)
